from flask import Flask, request, Response, make_response
import inspect
from typing import Callable, Dict, Optional, Tuple, Type

from smartoutlet import ALL_OUTLET_CLASSES, OutletInterface


app = Flask(__name__)


class InvalidOutletException(Exception):
    pass


def create_arg_map(outlettype: str) -> Tuple[Type[OutletInterface], Dict[str, Tuple[Callable[[Optional[str]], object], str]]]:
    outmap: Dict[str, Tuple[Callable[[Optional[str]], object], str]] = {}

    for clz in ALL_OUTLET_CLASSES:
        if clz.type.lower() == outlettype.lower():
            # Figure out arguments to add for this outlet.
            signature = inspect.signature(clz.__init__)
            for param in signature.parameters.values():
                if param.name == "self":
                    continue
                if param.default is inspect.Parameter.empty:
                    def requiredcreator(annotation: Callable[[str], object]) -> Callable[[Optional[str]], object]:
                        def cons(arg: Optional[str]) -> object:
                            if arg is None:
                                raise TypeError(f"Expected {annotation.__name__}, got None!")
                            return annotation(arg)
                        return cons

                    outmap[param.name] = requiredcreator(param.annotation), param.annotation.__name__
                else:
                    def defaultcreator(annotation: Callable[[str], object], default: object) -> Callable[[Optional[str]], object]:
                        def cons(arg: Optional[str]) -> object:
                            if arg is None:
                                return default
                            return annotation(arg)
                        return cons

                    outmap[param.name] = defaultcreator(param.annotation, param.default), param.annotation.__name__
            return clz, outmap

    raise InvalidOutletException(f"Unrecognized outlet type {outlettype}!")


@app.route('/<outlettype>', methods=['GET'])
def query_outlet(outlettype: str) -> Response:
    try:
        clz, args = create_arg_map(outlettype)
    except InvalidOutletException as e:
        return make_response(str(e), 400)

    argmap: Dict[str, object] = {}
    for k, (cons, objtype) in args.items():
        try:
            argmap[k] = cons(request.args.get(k))
        except TypeError:
            return make_response(f"Outlet type {outlettype} requires parameter {k} to be of type {objtype}!")

    try:
        inst = clz.deserialize(argmap)
        state = inst.getState()
        if state is None:
            resp = "unknown"
        else:
            resp = "on" if state else "off"

        return make_response(resp, 200)
    except Exception as e:
        return make_response(str(e), 400)


@app.route('/<outlettype>', methods=['PUT', 'POST', 'PATCH'])
def update_outlet(outlettype: str) -> Response:
    try:
        clz, args = create_arg_map(outlettype)
    except InvalidOutletException as e:
        return make_response(str(e), 400)

    data = request.data.decode('utf-8')
    state = None
    if data.lower() == "on":
        state = True
    elif data.lower() == "off":
        state = False
    else:
        return make_response("Request body should be either \"on\" or \"off\"", 400)

    argmap: Dict[str, object] = {}
    for k, (cons, objtype) in args.items():
        try:
            argmap[k] = cons(request.args.get(k))
        except TypeError:
            return make_response(f"Outlet type {outlettype} requires parameter {k} to be of type {objtype}!")

    try:
        inst = clz.deserialize(argmap)
        inst.setState(state)
        state = inst.getState()
        if state is None:
            resp = "unknown"
        else:
            resp = "on" if state else "off"

        return make_response(resp, 200)
    except Exception as e:
        return make_response(str(e), 400)



from typing import ClassVar, Dict, Optional, cast

from .interface import OutletInterface, param


@param("host", "the hostname or IP address of the NP-02 you are attempting to control")
@param("outlet", "the outlet number (between 1-2 inclusive) that you are attempting to control")
@param("community", "the SNMP read/write community as specified in the NP-02 config menu")
class NP02Outlet(OutletInterface):
    type: ClassVar[str] = "np-02"

    def __init__(
        self,
        *,
        host: str,
        outlet: int,
        # Yes, they only support one community for read and write.
        community: str = "public",
    ) -> None:
        self.host = host
        self.outlet = outlet
        self.community = community

        # Import these here to pay less cost to startup time.
        import pysnmp.hlapi as snmplib  # type: ignore
        import pysnmp.proto.rfc1902 as rfc1902  # type: ignore
        self.snmplib = snmplib
        self.rfc1902 = rfc1902

    def serialize(self) -> Dict[str, object]:
        return {
            'host': self.host,
            'outlet': self.outlet,
            'community': self.community,
        }

    @staticmethod
    def deserialize(vals: Dict[str, object]) -> OutletInterface:
        return NP02Outlet(
            host=cast(str, vals['host']),
            outlet=cast(int, vals['outlet']),
            community=cast(str, vals['community']),
        )

    def query(self, value: object) -> Optional[int]:
        try:
            return int(str(value))
        except ValueError:
            return None

    def update(self, value: bool) -> object:
        return self.rfc1902.Integer(1 if value else 2)

    def getState(self) -> Optional[bool]:
        iterator = self.snmplib.getCmd(
            self.snmplib.SnmpEngine(),
            self.snmplib.CommunityData(self.community, mpModel=0),
            self.snmplib.UdpTransportTarget((self.host, 161), timeout=1.0, retries=0),
            self.snmplib.ContextData(),
            self.snmplib.ObjectType(self.snmplib.ObjectIdentity(f"1.3.6.1.4.1.21728.2.4.1.2.1.1.3.{self.outlet}")),
        )

        for response in iterator:
            errorIndication, errorStatus, errorIndex, varBinds = response
            if errorIndication:
                return None
            elif errorStatus:
                return None
            else:
                for varBind in varBinds:
                    actual = self.query(varBind[1])

                    # Yes, this is the documented response, they clearly had a bug
                    # where they couldn't clear the top bit so the outlets modify
                    # each other and they just documented it as such.
                    if actual in {0, 256, 2}:
                        return False
                    elif actual in {1, 257}:
                        return True
                    return None
        return None

    def setState(self, state: bool) -> None:
        iterator = self.snmplib.setCmd(
            self.snmplib.SnmpEngine(),
            self.snmplib.CommunityData(self.community, mpModel=0),
            self.snmplib.UdpTransportTarget((self.host, 161)),
            self.snmplib.ContextData(),
            self.snmplib.ObjectType(self.snmplib.ObjectIdentity(f"1.3.6.1.4.1.21728.2.4.1.2.1.1.4.{self.outlet}"), self.update(state)),
        )
        next(iterator)



from typing import ClassVar, Dict, Optional, cast

from .interface import OutletInterface, param


@param("host", "the hostname or IP address of the NP-02B you are attempting to control")
@param("outlet", "the outlet number (between 1-2 inclusive) that you are attempting to control")
@param("username", "the administrator username as specified in the NP-02B web interface")
@param("password", "the administrator password as specified in the NP-02B web interface")
class NP02BOutlet(OutletInterface):
    type: ClassVar[str] = "np-02b"

    def __init__(
        self,
        *,
        host: str,
        outlet: int,
        username: str = "admin",
        password: str = "admin",
    ) -> None:
        self.host = host
        self.outlet = outlet
        self.username = username
        self.password = password

        # Import these here to pay less cost in import time.
        import requests
        import xml.etree.ElementTree as ET
        self.requests = requests
        self.ET = ET

    def serialize(self) -> Dict[str, object]:
        return {
            'host': self.host,
            'outlet': self.outlet,
            'username': self.username,
            'password': self.password,
        }

    @staticmethod
    def deserialize(vals: Dict[str, object]) -> OutletInterface:
        return NP02BOutlet(
            host=cast(str, vals['host']),
            outlet=cast(int, vals['outlet']),
            username=cast(str, vals['username']),
            password=cast(str, vals['password']),
        )

    def getState(self, force_legacy: bool = False) -> Optional[bool]:
        # We allow a force-legacy option here, because we call getState from within
        # setState, and if we have to call this we already know that it's a legacy
        # NP-02B. So, stop wasting time figuring that out a second time!
        if not force_legacy:
            try:
                response = self.requests.get(
                    f"http://{self.username}:{self.password}@{self.host}/cmd.cgi?$A5",
                    timeout=1.0,
                ).content.decode('utf-8').strip()
            except (self.requests.exceptions.ConnectTimeout, self.requests.exceptions.ConnectionError):
                return None
        else:
            # Shouldn't ever get to the bottom stanza, but lets be sure anyway.
            response = "$"

        # There are two types of response here, if it returns "Success!" then
        # it doesn't respond to the correct documented protocol.
        if force_legacy or response == "Success!":
            relay = f"rly{self.outlet - 1}"
            response = self.requests.get(
                f"http://{self.username}:{self.password}@{self.host}/status.xml",
                timeout=1.0
            ).content.decode('utf-8')

            root = self.ET.fromstring(response)
            if root.tag == "response":
                for child in root:
                    if child.tag == relay:
                        return child.text != '0'
            return None

        if '$' in response:
            return None
        if len(response) < (self.outlet - 1):
            return None
        return response[-self.outlet] != '0'

    def setState(self, state: bool) -> None:
        try:
            response = self.requests.get(
                f"http://{self.username}:{self.password}@{self.host}/cmd.cgi?$A3 {self.outlet} {'1' if state else '0'}",
                timeout=1.0
            ).content.decode('utf-8').strip()
        except (self.requests.exceptions.ConnectTimeout, self.requests.exceptions.ConnectionError):
            return

        if response == "Success!":
            # This outlet is not responding to the correct documented protocol,
            # we must query the status and then flip the relay if needed.
            actual = self.getState(force_legacy=True)
            if actual is None:
                # Couldn't query, so don't want to mess with toggling the relay.
                return

            if actual != state:
                try:
                    # Need to toggle
                    self.requests.get(f"http://{self.username}:{self.password}@{self.host}/cmd.cgi?rly={self.outlet - 1}", timeout=1.0)
                except (self.requests.exceptions.ConnectTimeout, self.requests.exceptions.ConnectionError):
                    pass



import logging
import os
import pickle
import socket
import struct
import sys
import threading
import time
from typing import ClassVar, Dict, Final, List, Optional, cast

from . import ALL_OUTLET_CLASSES
from .interface import OutletInterface


PROXY_VERSION: Final[int] = 5
PROXY_PORT: Final[int] = 54545
PROXY_CACHE_TIME: Final[float] = 5.0
PROXY_CACHE_REFILL_TIME: Final[float] = 3.0
PROXY_POOL_SIZE: Final[int] = 10


exit_daemon: bool = False


class OutletConnection:
    def __init__(self, port: int) -> None:
        self.port: int = port
        self.sock: Optional[socket.socket] = None

    def send(self, data: object) -> None:
        if self.sock is None:
            return
        pickled = pickle.dumps(data)
        self.sock.send(struct.pack("<I", len(pickled)) + pickled)

    def recv(self) -> object:
        if self.sock is None:
            return None

        data = self.sock.recv(4)
        length = struct.unpack("<I", data)[0]

        data = b""
        while len(data) < length:
            data += self.sock.recv(length)

        return pickle.loads(data)

    def connect(self) -> bool:
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect(("localhost", self.port))
        except ConnectionRefusedError:
            return False

        self.sock = sock
        self.send(("checkVersion", PROXY_VERSION))
        retval = self.recv()
        if isinstance(retval, bool):
            return retval
        return False

    def disconnect(self) -> None:
        if self.sock is not None:
            self.sock.close()
            self.sock = None


class OutletProxy(OutletInterface):
    type: ClassVar[str] = "proxy"

    def __init__(self, conn: OutletConnection, vals: Dict[str, object]) -> None:
        self.vals = vals
        self.conn = conn

    def serialize(self) -> Dict[str, object]:
        # We don't implement this, because this is a local proxy object only.
        raise NotImplementedError("Do not serialize proxy outlets!")

    @staticmethod
    def __connect(port: int) -> Optional[OutletConnection]:
        conn = OutletConnection(port)
        if conn.connect():
            return conn
        else:
            return None

    @staticmethod
    def deserialize(vals: Dict[str, object]) -> "OutletInterface":
        # We use this to connect to a remote interface
        if "type" not in vals:
            raise Exception("Could not instantiate a deserialization of an abstract outlet!")

        if "port" in vals:
            port = cast(int, vals['port'] or PROXY_PORT)
            del vals['port']
        else:
            port = PROXY_PORT

        logloc = None
        if "log" in vals:
            logloc = cast(str, vals["log"])
            del vals["log"]

        # Attempt to connect to an existing remote daemon that's already started.
        proxy = OutletProxy.__connect(port)

        # If it is not already running, attempt to start a new one.
        if proxy is None:
            pid = os.fork()
            if pid == 0:
                # Decouple from parent.
                os.chdir("/")
                os.setsid()
                os.umask(0)

                # Secondary fork.
                pid = os.fork()
                if pid > 0:
                    # We're the parent, we should exit.
                    sys.exit(0)

                # Set up logging to go to file.
                if logloc is not None:
                    try:
                        os.remove(logloc)
                    except FileNotFoundError:
                        pass
                    logging.basicConfig(filename=logloc, level=logging.INFO)

                # Now, start the server daemon.
                for _ in range(500):
                    try:
                        daemon = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                        daemon.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                        daemon.settimeout(0.05)
                        daemon.bind(("localhost", port))
                        daemon.listen(1)

                        logging.info(f"Started daemon server listening on {port}")
                        break
                    except OSError:
                        # Can happen when restarting server.
                        time.sleep(0.01)
                else:
                    raise Exception("Failed to spawn proxy daemon instance!")

                # Now, run the loop until we're requested to exit.
                outletdaemon = OutletDaemon()
                while not exit_daemon:
                    try:
                        conn, _ = daemon.accept()
                    except socket.timeout:
                        time.sleep(0.01)
                        continue

                    def reply(data: object) -> None:
                        pickled = pickle.dumps(data)
                        conn.send(struct.pack("<I", len(pickled)) + pickled)

                    try:
                        while True:
                            # Handle the message
                            data = conn.recv(4)
                            length = struct.unpack("<I", data)[0]

                            data = b""
                            while len(data) < length:
                                data += conn.recv(length)

                            actual = pickle.loads(data)
                            if isinstance(actual, tuple) and len(actual) == 2:
                                cmd, param = actual
                                if cmd == "checkVersion":
                                    reply(outletdaemon.checkVersion(param))
                                elif cmd == "getState":
                                    reply(outletdaemon.getState(param))
                                elif cmd == "setState":
                                    outletdaemon.setState(*param)
                                else:
                                    logging.warning(f"Unrecognized command {cmd}!")
                            else:
                                logging.warning("Unrecognized packet!")

                    except Exception:
                        pass
                    finally:
                        conn.close()

                sys.exit(0)

            else:
                for _ in range(500):
                    try:
                        proxy = OutletProxy.__connect(port)
                    except OSError:
                        # Can happen when restarting server.
                        proxy = None

                    if proxy is not None:
                        break
                    time.sleep(0.01)
                else:
                    raise Exception("Failed to spawn proxy daemon instance!")

        return OutletProxy(proxy, vals)

    def getState(self) -> Optional[bool]:
        # This is where we talk to the remote daemon.
        self.conn.send(("getState", self.vals))
        return cast(Optional[bool], self.conn.recv())

    def setState(self, state: bool) -> None:
        # This is where we talk to the remote daemon.
        self.conn.send(("setState", (self.vals, state)))


def pollThread(daemon: "OutletDaemon", num: int) -> None:
    daemon.poll(num)


class OutletDaemon:
    def __init__(self) -> None:
        # Cache contrl and registered outlets.
        self.registered_outlets: Dict[str, OutletInterface] = {}
        self.cached_states: Dict[str, Optional[bool]] = {}
        self.cached_times: Dict[str, float] = {}

        # Threadpool control structures
        self.pending_keys_lock: threading.Lock = threading.Lock()
        self.pending_keys: List[str] = []
        self.thread_count: int = 0

    def start_thread(self) -> None:
        thread = None
        with self.pending_keys_lock:
            if self.thread_count < PROXY_POOL_SIZE:
                self.thread_count += 1
                thread = threading.Thread(target=pollThread, args=(self, self.thread_count), daemon=True)

        if thread is not None:
            thread.start()

    def poll(self, num: int) -> None:
        logging.debug(f"Starting polling thread {num}")

        while True:
            try:
                # First, let's do some bookkeeping.
                with self.pending_keys_lock:
                    if not self.pending_keys:
                        # We're clever to use cached times here, so the initial poll doesn't
                        # cause us to make the CLI wait longer.
                        logging.debug("Full sweep completed, starting over!")
                        self.pending_keys = [x for x in self.cached_times]

                    key = None
                    if self.pending_keys:
                        key = self.pending_keys.pop()

                if key is not None and self.cached_times[key] < (time.time() - PROXY_CACHE_REFILL_TIME):
                    # Now, let's fetch the outlet in order to refresh.
                    outlet = self.registered_outlets[key]
                    logging.debug(f"Caching state for {key}")
                    self.cached_states[key] = outlet.getState()
                    self.cached_times[key] = time.time()
                    logging.debug(f"Cached state for {key} is {self.cached_states[key]}")
                else:
                    # Sleep a little longer, given we had nothing to do.
                    time.sleep(0.5)

            except Exception:
                logging.exception(f"Failed to poll in thread {num}")

            # Sleep a little, give the rest of the code some breathing room.
            time.sleep(0.05)

    def checkVersion(self, proxy_version: int) -> bool:
        if proxy_version == PROXY_VERSION:
            return True

        # We need to kill ourselves, we're running the wrong version!
        logging.info("We are running the wrong version, so time to die!")

        global exit_daemon
        exit_daemon = True

        return False

    def __getKey(self, vals: Dict[str, object]) -> str:
        return "-".join(f"{k}:{vals[k]}" for k in sorted(vals.keys()))

    def __getClass(self, vals: Dict[str, object]) -> OutletInterface:
        key = self.__getKey(vals)
        knowntype: str = cast(str, vals['type'])
        del vals['type']

        if key not in self.registered_outlets:
            for clz in ALL_OUTLET_CLASSES:
                if clz.type.lower() == knowntype.lower():
                    logging.info(f"Registering new outlet with key {key}")
                    self.registered_outlets[key] = clz.deserialize(vals)
                    self.start_thread()

                    break
            else:
                raise Exception(f"Cannot deserialize outlet of type {knowntype}!")

        return self.registered_outlets[key]

    def getState(self, vals: Dict[str, object]) -> Optional[bool]:
        key = self.__getKey(vals)
        if key not in self.cached_states or self.cached_times[key] < (time.time() - PROXY_CACHE_TIME):
            outlet = self.__getClass(vals)
            logging.info(f"Fetching state for {key}")
            self.cached_states[key] = outlet.getState()
            self.cached_times[key] = time.time()

        logging.info(f"State for {key} is {self.cached_states[key]}")
        return self.cached_states[key]

    def setState(self, vals: Dict[str, object], state: bool) -> None:
        key = self.__getKey(vals)
        outlet = self.__getClass(vals)
        logging.info(f"Setting state for {key} to {state}")
        outlet.setState(state)
        self.cached_states[key] = outlet.getState()
        self.cached_times[key] = time.time()
        logging.info(f"State for {key} is {self.cached_states[key]}")



from typing import Dict, Optional, Type, cast

from .interface import OutletInterface


class AP790XOutlet(OutletInterface):
    def __init__(self, *, host: str, outlet: int, read_community: str = "public", write_community: str = "private") -> None:
        if outlet < 1 or outlet > 8:
            raise Exception("Out of bounds outlet number!")

        # Import this here to pay less cost to speed of startup.
        from .snmp import SNMPOutlet

        self.host = host
        self.outlet = outlet
        self.read_community = read_community
        self.write_community = write_community
        self.snmp = SNMPOutlet(
            host=host,
            query_oid=f"1.3.6.1.4.1.318.1.1.12.3.3.1.1.4.{outlet}",
            query_on_value=1,
            query_off_value=2,
            update_oid=f"1.3.6.1.4.1.318.1.1.12.3.3.1.1.4.{outlet}",
            update_on_value=1,
            update_off_value=2,
            read_community=read_community,
            write_community=write_community,
        )

    def serialize(self) -> Dict[str, object]:
        return {
            'host': self.host,
            'outlet': self.outlet,
            'read_community': self.read_community,
            'write_community': self.write_community,
        }

    @classmethod
    def deserialize(cls: "Type[AP790XOutlet]", vals: Dict[str, object]) -> "AP790XOutlet":
        return cls(
            host=cast(str, vals['host']),
            outlet=cast(int, vals['outlet']),
            read_community=cast(str, vals['read_community']),
            write_community=cast(str, vals['write_community']),
        )

    def getState(self) -> Optional[bool]:
        return self.snmp.getState()

    def setState(self, state: bool) -> None:
        return self.snmp.setState(state)



from .interface import OutletInterface
from .snmp import SNMPOutlet
from .ap7900 import AP7900Outlet
from .ap7901 import AP7901Outlet
from .np_02 import NP02Outlet
from .np_02b import NP02BOutlet

from typing import List, Type


ALL_OUTLET_CLASSES: List[Type[OutletInterface]] = [SNMPOutlet, AP7900Outlet, AP7901Outlet, NP02Outlet, NP02BOutlet]

__all__ = [
    "OutletInterface",
    "SNMPOutlet",
    "AP7900Outlet",
    "AP7901Outlet",
    "NP02Outlet",
    "NP02BOutlet",
    "ALL_OUTLET_CLASSES",
]



from typing import ClassVar

from .interface import param
from .ap790x import AP790XOutlet


@param("host", "the hostname or IP address of the AP7900 you are attempting to control")
@param("outlet", "the outlet number (between 1-8 inclusive) that you are attempting to control")
@param("read_community", "the SNMP read community as specified in the AP7900 config menu")
@param("write_community", "the SNMP write community as specified in the AP7900 config menu")
class AP7900Outlet(AP790XOutlet):
    type: ClassVar[str] = "ap7900"



from abc import ABC, abstractmethod
from typing import Any, Callable, ClassVar, Dict, Optional, TypeVar


class OutletInterface(ABC):
    type: ClassVar[str]

    @abstractmethod
    def serialize(self) -> Dict[str, object]:
        ...

    @staticmethod
    @abstractmethod
    def deserialize(vals: Dict[str, object]) -> "OutletInterface":
        ...

    @abstractmethod
    def getState(self) -> Optional[bool]:
        ...

    @abstractmethod
    def setState(self, state: bool) -> None:
        ...


F = TypeVar('F', bound=Callable[..., Any])


def param(param: str, description: str) -> Callable[[F], F]:
    def decorator(clz: F) -> F:
        paramdict = getattr(clz, "__params__", {})
        paramdict[param] = description
        setattr(clz, "__params__", paramdict)
        return clz

    return decorator
